@@ -1,10 +1,14 @@
 # Minimal support for package require
 # No error on failure since C extensions aren't handled
-proc package {cmd pkg} {
+proc package {cmd pkg args} {
 	if {$cmd eq "require"} {
 		foreach path $::auto_path {
-			if {[file exists $path/$pkg.tcl]} {
-				uplevel #0 [list source $path/$pkg.tcl]
+			set pkgpath $path/$pkg.tcl
+			if {$path eq "."} {
+				set pkgpath $pkg.tcl
+			}
+			if {[file exists $pkgpath]} {
+				uplevel #0 [list source $pkgpath]
 				return
 			}
 		}
