@@ -1,5 +1,12 @@
 # Implements script-based standard commands for Jim Tcl
 
+if {![exists -command ref]} {
+	# No support for references, so create a poor-man's reference just good enough for lambda
+	proc ref {args} {{count 0}} {
+		format %08x [incr count]
+	}
+}
+
 # Creates an anonymous procedure
 proc lambda {arglist args} {
 	tailcall proc [ref {} function lambda.finalizer] $arglist {*}$args
@@ -59,6 +66,13 @@ proc stackdump {stacktrace} {
 	join $lines \n
 }
 
+# Add the given script to $jim::defer, to be evaluated when the current
+# procedure exits
+proc defer {script} {
+	upvar jim::defer v
+	lappend v $script
+}
+
 # Sort of replacement for $::errorInfo
 # Usage: errorInfo error ?stacktrace?
 proc errorInfo {msg {stacktrace ""}} {
@@ -87,27 +101,6 @@ proc {info nameofexecutable} {} {
 	}
 }
 
-# Script-based implementation of 'dict with'
-proc {dict with} {&dictVar {args key} script} {
-	set keys {}
-	foreach {n v} [dict get $dictVar {*}$key] {
-		upvar $n var_$n
-		set var_$n $v
-		lappend keys $n
-	}
-	catch {uplevel 1 $script} msg opts
-	if {[info exists dictVar] && ([llength $key] == 0 || [dict exists $dictVar {*}$key])} {
-		foreach n $keys {
-			if {[info exists var_$n]} {
-				dict set dictVar {*}$key $n [set var_$n]
-			} else {
-				dict unset dictVar {*}$key $n
-			}
-		}
-	}
-	return {*}$opts $msg
-}
-
 # Script-based implementation of 'dict update'
 proc {dict update} {&varName args script} {
 	set keys {}
@@ -130,19 +123,6 @@ proc {dict update} {&varName args script} {
 	return {*}$opts $msg
 }
 
-# Script-based implementation of 'dict merge'
-# This won't get called in the trivial case of no args
-proc {dict merge} {dict args} {
-	foreach d $args {
-		# Check for a valid dict
-		dict size $d
-		foreach {k v} $d {
-			dict set dict $k $v
-		}
-	}
-	return $dict
-}
-
 proc {dict replace} {dictionary {args {key value}}} {
 	if {[llength ${key value}] % 2} {
 		tailcall {dict replace}
@@ -188,11 +168,6 @@ proc {dict remove} {dictionary {args key}} {
 	return $dictionary
 }
 
-# Script-based implementation of 'dict values'
-proc {dict values} {dictionary {pattern *}} {
-	dict keys [lreverse $dictionary] $pattern
-}
-
 # Script-based implementation of 'dict for'
 proc {dict for} {vars dictionary script} {
 	if {[llength $vars] != 2} {
